"""
    loginpass.stackapps
    ~~~~~~~~~~~~~~~~~~~

    Authenticate using `Stack Overflow <https://stackoverflow.com>`_.

    Useful links:

    - Register an application: https://stackapps.com/apps/oauth/register
    - API documentation: https://api.stackexchange.com/

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

from authlib.specs.oidc import UserInfo
from authlib.common.urls import add_params_to_uri, url_decode
from ._core import OAuthBackend, map_profile_fields


def _stackapps_compliance_fix(session, site):
    def _add_extra_info(url, headers, body):
        params = {'site': site}
        api_key = session._kwargs.get('api_key')
        api_filter = session._kwargs.get('api_filter')
        if api_key:
            params['key'] = api_key
        if api_filter:
            params['filter'] = api_filter
        url = add_params_to_uri(url, params)
        return url, headers, body

    def _token_response(resp):
        data = dict(url_decode(resp.text))
        data['token_type'] = 'Bearer'
        data['expires_in'] = int(data['expires'])
        resp.json = lambda: data
        return resp

    session.register_compliance_hook('protected_request', _add_extra_info)
    session.register_compliance_hook('access_token_response', _token_response)


def create_stackapps_backend(name, site):
    """Build StackApps OAuth Backend."""

    authorize_url = 'https://stackoverflow.com/oauth'
    token_url = 'https://stackoverflow.com/oauth/access_token'
    api_base_url = 'https://api.stackexchange.com/2.2'

    class StackApp(OAuthBackend):
        OAUTH_TYPE = '2.0'
        OAUTH_NAME = name
        OAUTH_CONFIG = {
            'api_base_url': api_base_url,
            'access_token_url': token_url,
            'authorize_url': authorize_url,
            'client_kwargs': {
                'token_placement': 'uri',
                'token_endpoint_auth_method': 'client_secret_post',
            },
            'compliance_fix': lambda s: _stackapps_compliance_fix(s, site)
        }
        # you can add api_key, api_filter in client_kwargs

        def profile(self, **kwargs):
            """Get the user's profile."""
            resp = self.get('me', **kwargs)
            resp.raise_for_status()
            data = resp.json()
            return UserInfo(map_profile_fields(data['items'][0], {
                'sub': lambda o: str(o['user_id']),
                'name': 'display_name',
                'preferred_username': 'display_name',
                'profile': 'link',
                'picture': 'profile_image',
                'website': 'website_url',
                'address': 'location',
                'updated_at': 'last_modified_date',
            }))

    return StackApp


StackOverflow = create_stackapps_backend('stackoverflow', 'stackoverflow')



"""
    loginpass.discord
    ~~~~~~~~~~~~~~~~~

    Loginpass Backend of Discord (https://discordapp.com).

    Useful Links:

    - Create App: https://discordapp.com/developers/applications/me
    - API documentation: https://discordapp.com/developers/docs/reference

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend, map_profile_fields
from ._consts import version, homepage

# see: https://discordapp.com/developers/docs/reference#user-agent
UA = 'Loginpass ({}, {})'.format(homepage, version)


class Discord(OAuthBackend):
    OAUTH_TYPE = '2.0'
    OAUTH_NAME = 'discord'
    OAUTH_CONFIG = {
        'api_base_url': 'https://discordapp.com/api/',
        'access_token_url': 'https://discordapp.com/api/oauth2/token',
        'authorize_url': 'https://discordapp.com/api/oauth2/authorize',
        'client_kwargs': {
            'token_endpoint_auth_method': 'client_secret_post',
            'scope': 'identify email'
        },
    }
    DEFAULT_USER_AGENT = UA

    def profile(self, **kwargs):
        resp = self.get('users/%40me', **kwargs)
        resp.raise_for_status()
        data = resp.json()
        params = map_profile_fields(data, {
            'sub': 'id',
            'name': 'username',
            'email': 'email',
            'preferred_username': 'username',
            'email_verified': 'verified',
        })
        if 'avatar' in data:
            src = 'https://cdn.discordapp.com/avatars/{}/{}.png'
            params['picture'] = src.format(data['id'], data['avatar'])
        return UserInfo(params)



"""
    loginpass.dropbox
    ~~~~~~~~~~~~~~~~~

    Loginpass Backend of Dropbox (https://dropbox.com).

    Useful Links:

    - Create App: https://www.dropbox.com/developers/apps
    - API documentation: https://www.dropbox.com/developers

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend


class Dropbox(OAuthBackend):
    OAUTH_TYPE = '2.0'
    OAUTH_NAME = 'dropbox'
    OAUTH_CONFIG = {
        'api_base_url': 'https://www.dropbox.com/2/',
        'access_token_url': 'https://api.dropboxapi.com/oauth2/token',
        'authorize_url': 'https://www.dropbox.com/oauth2/authorize',
    }

    def profile(self, **kwargs):
        resp = self.post('users/get_current_account', **kwargs)
        resp.raise_for_status()
        data = resp.json()
        name_info = data['name']
        params = {
            'sub': data['account_id'],
            'name': name_info.get('display_name'),
            'given_name': name_info.get('given_name'),
            'family_name': name_info.get('surname'),
            'nickname': name_info.get('familiar_name'),
            'email': data.get('email'),
            'email_verified': data.get('email_verified'),
            'locale': data.get('locale'),
            'picture': data.get('profile_photo_url'),
        }
        return UserInfo(params)



"""
    loginpass.facebook
    ~~~~~~~~~~~~~~~~~~

    Loginpass Backend of Facebook (https://facebook.com).

    Useful Links:

    - Create App: https://developers.facebook.com/apps
    - API documentation: https://developers.facebook.com/docs/graph-api

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend, map_profile_fields


class Facebook(OAuthBackend):
    OAUTH_TYPE = '2.0'
    OAUTH_NAME = 'facebook'
    OAUTH_CONFIG = {
        'api_base_url': 'https://graph.facebook.com/v2.12',
        'access_token_url': 'https://graph.facebook.com/v2.12/oauth/access_token',
        'authorize_url': 'https://www.facebook.com/v2.12/dialog/oauth',
        'client_kwargs': {'scope': 'email public_profile'},
    }

    def profile(self, **kwargs):
        if 'params' not in kwargs:
            fields = [
                'id', 'name', 'first_name', 'middle_name', 'last_name',
                'email', 'website', 'gender', 'locale'
            ]
            kwargs['params'] = {'fields': ','.join(fields)}
        resp = self.get('me', **kwargs)
        resp.raise_for_status()
        return UserInfo(map_profile_fields(resp.json(), {
            'sub': lambda o: str(o['id']),
            'name': 'name',
            'given_name': 'first_name',
            'family_name': 'last_name',
            'middle_name': 'middle_name',
            'email': 'email',
            'website': 'website',
            'gender': 'gender',
            'locale': 'locale'
        }))



"""
    loginpass.auth0
    ~~~~~~~~~~~~~~~

    Loginpass Backend of Auth0 (https://auth0.com) tenants'
    endpoints.

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend, parse_id_token

ALLOW_REGIONS = ['us', 'eu', 'au']


def create_auth0_backend(name, tenant, region=None):
    """Create an Auth0 backend for one tenant."""

    if region and region not in ALLOW_REGIONS:
        raise ValueError('Not a vaild "region"')
    if not region or region == 'us':
        host = 'https://{}.auth0.com/'.format(tenant)
    else:
        host = 'https://{}.{}.auth0.com/'.format(tenant, region)

    authorize_url = '{}authorize'.format(host)
    token_url = '{}oauth/token'.format(host)

    class Auth0(OAuthBackend):
        OAUTH_TYPE = '2.0,oidc'
        OAUTH_NAME = name
        OAUTH_CONFIG = {
            'api_base_url': host,
            'access_token_url': token_url,
            'authorize_url': authorize_url,
            'client_kwargs': {'scope': 'openid email profile'},
        }
        JWK_SET_URL = host + '.well-known/jwks.json'

        def profile(self, **kwargs):
            resp = self.get('userinfo', **kwargs)
            resp.raise_for_status()
            data = resp.json()
            return UserInfo(data)

        def parse_openid(self, token, nonce=None):
            return parse_id_token(
                self, token['id_token'],
                {"iss": {"values": [host]}},
                token.get('access_token'), nonce
            )

    return Auth0



"""
    loginpass.bitbucket
    ~~~~~~~~~~~~~~~~~~~

    Loginpass Backend of Bitbucket (https://bitbucket.org/).

    Useful Links:

    - Create App: https://bitbucket.org/account/
    - API documentation: https://developer.atlassian.com/cloud/bitbucket/

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend, map_profile_fields


class Bitbucket(OAuthBackend):
    OAUTH_TYPE = '2.0'
    OAUTH_NAME = 'bitbucket'
    OAUTH_CONFIG = {
        'api_base_url': 'https://api.bitbucket.org/2.0/',
        'access_token_url': 'https://bitbucket.org/site/oauth2/access_token',
        'authorize_url': 'https://bitbucket.org/site/oauth2/authorize',
        'client_kwargs': {'scope': 'email'},
    }

    def profile(self, **kwargs):
        resp = self.get('user', **kwargs)
        resp.raise_for_status()
        params = map_profile_fields(resp.json(), {
            'sub': 'account_id',
            'name': 'display_name',
            'preferred_username': 'username',
            'address': 'location',
            'website': 'website',
            'picture': _get_avatar,
            'profile': _get_profile,
        })
        resp = self.get('user/emails', **kwargs)
        resp.raise_for_status()
        params.update(_get_email(resp.json()))
        return UserInfo(params)


def _get_profile(data):
    return 'https://bitbucket.org/{}/'.format(data['username'])


def _get_avatar(data):
    avatar = data['links'].get('avatar')
    if avatar:
        return avatar.get('href')


def _get_email(data):
    confirmed_emails = []
    values = data['values']
    for value in values:
        if value['is_primary']:
            return {
                'email': value['email'],
                'email_verified': value['is_confirmed']
            }
        if value['is_confirmed']:
            confirmed_emails.append(value['email'])

    if confirmed_emails:
        return {'email': confirmed_emails[0], 'email_verified': True}

    if values:
        return {'email': values[0], 'email_verified': False}

    return {}



"""
    loginpass.google
    ~~~~~~~~~~~~~~~~

    This module contains a loginpass backend of Google,
    and a ServiceAccount requests Session.

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

import json
from authlib.client import AssertionSession
from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend, parse_id_token


GOOGLE_API_URL = 'https://www.googleapis.com/'
GOOGLE_TOKEN_URL = GOOGLE_API_URL + 'oauth2/v4/token'
GOOGLE_JWK_URL = GOOGLE_API_URL + 'oauth2/v3/certs'
GOOGLE_AUTH_URL = (
    'https://accounts.google.com/o/oauth2/v2/auth'
    '?access_type=offline'
)
GOOGLE_CLAIMS_OPTIONS = {
    "iss": {
        "values": ['https://accounts.google.com', 'accounts.google.com']
    }
}


class Google(OAuthBackend):
    OAUTH_TYPE = '2.0,oidc'
    OAUTH_NAME = 'google'
    OAUTH_CONFIG = {
        'api_base_url': GOOGLE_API_URL,
        'access_token_url': GOOGLE_TOKEN_URL,
        'authorize_url': GOOGLE_AUTH_URL,
        'client_kwargs': {'scope': 'openid email profile'},
    }
    JWK_SET_URL = GOOGLE_JWK_URL

    def profile(self, **kwargs):
        resp = self.get('oauth2/v3/userinfo', **kwargs)
        resp.raise_for_status()
        return UserInfo(resp.json())

    def parse_openid(self, token, nonce=None):
        return parse_id_token(
            self, token['id_token'], GOOGLE_CLAIMS_OPTIONS,
            token.get('access_token'), nonce
        )


class GoogleServiceAccount(AssertionSession):
    @classmethod
    def from_service_account_file(cls, conf_file, scope, subject=None):
        with open(conf_file, 'r') as f:
            conf = json.load(f)

        token_url = conf['token_uri']
        issuer = conf['client_email']
        key = conf['private_key']
        key_id = conf.get('private_key_id')

        header = {'alg': 'RS256'}
        if key_id:
            header['kid'] = key_id

        # Google puts scope in payload
        claims = {'scope': scope}
        return cls(
            grant_type=cls.JWT_BEARER_GRANT_TYPE,
            token_url=token_url,
            issuer=issuer,
            audience=token_url,
            claims=claims,
            subject=subject,
            key=key,
            header=header,
        )



"""
    loginpass.twitter
    ~~~~~~~~~~~~~~~~~

    Loginpass Backend of Twitter (https://twitter.com).

    Useful Links:

    - Create App: https://apps.twitter.com/
    - API documentation: https://developer.twitter.com/

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""
from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend


class Twitter(OAuthBackend):
    OAUTH_TYPE = '1.0'
    OAUTH_NAME = 'twitter'
    OAUTH_CONFIG = {
        'api_base_url': 'https://api.twitter.com/1.1/',
        'request_token_url': 'https://api.twitter.com/oauth/request_token',
        'access_token_url': 'https://api.twitter.com/oauth/access_token',
        'authorize_url': 'https://api.twitter.com/oauth/authenticate',
    }

    def profile(self, **kwargs):
        url = 'account/verify_credentials.json'
        if 'params' not in kwargs:
            kwargs['params'] = {
                'skip_status': True,
                'include_email': True
            }
        resp = self.get(url, **kwargs)
        resp.raise_for_status()
        data = resp.json()
        params = {
            'sub': data['id_str'],
            'name': data['name'],
            'email': data.get('email'),
            'locale': data.get('lang'),
            'picture': data.get('profile_image_url_https'),
            'preferred_username': data.get('screen_name'),
        }
        username = params['preferred_username']
        if username:
            params['profile'] = 'https://twitter.com/{}'.format(username)
        return UserInfo(params)



"""
    loginpass.azure
    ~~~~~~~~~~~~~~~

    Loginpass Backend of Azure AD.

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend, parse_id_token


def create_azure_backend(name, tenant):
    """Build Azure Active Directory OAuth Backend."""

    base_url = 'https://login.microsoftonline.com/'
    authorize_url = '{}{}/oauth2/authorize'.format(base_url, tenant)
    token_url = '{}{}/oauth2/token'.format(base_url, tenant)

    jwt_claims_options = {
        "iss": {
            "values": ['https://sts.windows.net/{}/'.format(tenant)]
        }
    }

    class AzureAD(OAuthBackend):
        OAUTH_TYPE = '2.0,oidc'
        OAUTH_NAME = name
        OAUTH_CONFIG = {
            'api_base_url': 'graph.microsoft.com',
            'access_token_url': token_url,
            'authorize_url': authorize_url,
            'client_kwargs': {'scope': 'openid email profile'},
        }
        JWK_SET_URL = '{}{}/discovery/keys'.format(base_url, tenant)

        def profile(self, **kwargs):
            url = '{}{}/openid/userinfo'.format(base_url, tenant)
            resp = self.get(url, **kwargs)
            resp.raise_for_status()
            return UserInfo(**resp.json())

        def parse_openid(self, token, nonce=None):
            return parse_id_token(
                self, token['id_token'], jwt_claims_options,
                token.get('access_token'), nonce
            )

    return AzureAD


Azure = create_azure_backend('azure', 'common')



from authlib.client import OAuthClient
from authlib.common.security import generate_token
from authlib.specs.rfc7519 import jwt, jwk
from authlib.specs.oidc import CodeIDToken, ImplicitIDToken, UserInfo


class OAuthBackend(OAuthClient):
    """Backend for OAuth Registry"""
    OAUTH_TYPE = None
    OAUTH_NAME = None
    OAUTH_CONFIG = None
    JWK_SET_URL = None

    def fetch_jwk_set(self, force=False):
        if not self.JWK_SET_URL:
            return None

        jwk_set = getattr(self, '_jwk_set', None)
        if jwk_set and not force:
            return jwk_set

        resp = self.get(self.JWK_SET_URL, withhold_token=True)
        self._jwk_set = resp.json()
        return self._jwk_set


def _get_oauth_client_cls(oauth):
    try:
        from authlib.flask.client import (
            OAuth as FlaskOAuth,
            RemoteApp as FlaskRemoteApp,
        )
        if isinstance(oauth, FlaskOAuth):
            return FlaskRemoteApp
    except ImportError:
        try:
            from authlib.django.client import (
                OAuth as DjangoOAuth,
                RemoteApp as DjangoRemoteApp,
            )
            if isinstance(oauth, DjangoOAuth):
                return DjangoRemoteApp
        except ImportError:
            pass


def register_to(backend, oauth, client_base=None):
    """Register a backend to OAuth instance.

    :param backend: An OAuthBackend
    :param oauth: Authlib OAuth instance
    :param client_base: This function will find a client_base
        automatically, Flask or Django.
    :return: backend instance
    """
    if client_base is None:
        client_base = _get_oauth_client_cls(oauth)

    config = backend.OAUTH_CONFIG.copy()
    if client_base:
        class RemoteApp(client_base, backend):
            pass
        config['client_cls'] = RemoteApp
    return oauth.register(backend.OAUTH_NAME, overwrite=True, **config)


def create_flask_blueprint(backend, oauth, handle_authorize):
    """Create a Flask blueprint that you can register it directly to Flask
    app. The blueprint contains two route: ``/auth`` and ``/login``::

        from flask import Flask
        from authlib.flask.client import OAuth
        from loginpass import create_flask_blueprint, GitHub

        app = Flask(__name__)
        oauth = OAuth(app)


        def handle_authorize(remote, token, user_info):
            if token:
                save_token(remote.name, token)
            if user_info:
                save_user(user_info)
                return user_page
            raise some_error

        github_bp = create_flask_blueprint(GitHub, oauth, handle_authorize)
        app.register_blueprint(github_bp, url_prefix='/github')

        # visit /github/login
        # callback /github/auth

    :param backend: An OAuthBackend
    :param oauth: Authlib Flask OAuth instance
    :param handle_authorize: A function to handle authorized response
    :return: Flask Blueprint instance
    """
    from flask import Blueprint, request, url_for, current_app, session
    from authlib.flask.client import RemoteApp

    remote = register_to(backend, oauth, RemoteApp)
    nonce_key = '_{}:nonce'.format(backend.OAUTH_NAME)
    bp = Blueprint('loginpass_' + backend.OAUTH_NAME, __name__)

    @bp.route('/auth')
    def auth():
        id_token = request.args.get('id_token')
        if request.args.get('code'):
            token = remote.authorize_access_token()
            if id_token:
                token['id_token'] = id_token
        elif id_token:
            token = {'id_token': id_token}
        else:
            # handle failed
            return handle_authorize(remote, None, None)
        if 'id_token' in token:
            nonce = session[nonce_key]
            user_info = remote.parse_openid(token, nonce)
        else:
            user_info = remote.profile()
        return handle_authorize(remote, token, user_info)

    @bp.route('/login')
    def login():
        redirect_uri = url_for('.auth', _external=True)
        conf_key = '{}_AUTHORIZE_PARAMS'.format(backend.OAUTH_NAME.upper())
        params = current_app.config.get(conf_key, {})
        if 'oidc' in backend.OAUTH_TYPE:
            nonce = generate_token(20)
            session[nonce_key] = nonce
            params['nonce'] = nonce
        return remote.authorize_redirect(redirect_uri, **params)

    return bp


def create_django_urlpatterns(backend, oauth, handle_authorize):
    from django.conf import settings
    from django.urls import path, reverse
    from authlib.django.client import RemoteApp

    config = getattr(settings, 'AUTHLIB_OAUTH_CLIENTS', None)
    authorize_params = None
    if config:
        backend_config = config.get(backend.OAUTH_NAME)
        if backend_config:
            authorize_params = backend_config.get('authorize_params')

    remote = register_to(backend, oauth, RemoteApp)
    nonce_key = '_{}:nonce'.format(backend.OAUTH_NAME)
    auth_route_name = 'loginpass_{}_auth'.format(backend.OAUTH_NAME)
    login_route_name = 'loginpass_{}_login'.format(backend.OAUTH_NAME)

    def auth(request):
        id_token = request.GET.get('id_token')
        if request.GET.get('code'):
            token = remote.authorize_access_token(request)
            if id_token:
                token['id_token'] = id_token
        elif id_token:
            token = {'id_token': id_token}
        else:
            # handle failed
            return handle_authorize(remote, None, None)
        if 'id_token' in token:
            nonce = request.session[nonce_key]
            user_info = remote.parse_openid(token, nonce)
        else:
            user_info = remote.profile(token=token)
        return handle_authorize(request, remote, token, user_info)

    def login(request):
        redirect_uri = request.build_absolute_uri(reverse(auth_route_name))
        params = {}
        if authorize_params:
            params.udpate(authorize_params)
        if 'oidc' in backend.OAUTH_TYPE:
            nonce = generate_token(20)
            request.session[nonce_key] = nonce
            params['nonce'] = nonce
        return remote.authorize_redirect(request, redirect_uri, **params)

    return [
        path('auth/', auth, name=auth_route_name),
        path('login/', login, name=login_route_name),
    ]


def map_profile_fields(data, fields):
    """Copy profile data from site-specific to standard field names.
    Standard keys will only be set if the site data for that key is not
    ``None``.

    :param data: Profile data from the site, to be modified in place.
    :param fields: Map of ``{destination: source}``. Destination is the
        standard name. Source source is the site-specific name, or
        a callable taking ``data`` and returning the value.
    :return: UserInfo fields
    """
    profile = {}
    for dst, src in fields.items():
        if callable(src):
            value = src(data)
        else:
            value = data.get(src)

        if value is not None:
            profile[dst] = value

    return profile


def parse_id_token(remote, id_token, claims_options,
                   access_token=None, nonce=None):
    """Parse UserInfo from id_token."""

    def load_key(header, payload):
        jwk_set = remote.fetch_jwk_set()
        try:
            return jwk.loads(jwk_set, header.get('kid'))
        except ValueError:
            jwk_set = remote.fetch_jwk_set(force=True)
            return jwk.loads(jwk_set, header.get('kid'))

    claims_params = dict(
        nonce=nonce,
        client_id=remote.client_id,
    )
    if access_token:
        claims_params['access_token'] = access_token
        claims_cls = CodeIDToken
    else:
        claims_cls = ImplicitIDToken
    claims = jwt.decode(
        id_token, key=load_key,
        claims_cls=claims_cls,
        claims_options=claims_options,
        claims_params=claims_params,
    )
    claims.validate(leeway=120)
    return UserInfo(claims)



"""
    loginpass.slack
    ~~~~~~~~~~~~~~~

    Loginpass Backend of Slack (https://slack.com).

    Useful Links:

    - Create App: https://api.slack.com/apps
    - API documentation: https://api.slack.com/
    - Sign In: https://api.slack.com/docs/sign-in-with-slack

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

import json
from authlib.specs.oidc import UserInfo
from authlib.common.encoding import to_unicode
from ._core import OAuthBackend


def slack_compliance_fix(session):
    def _fix(resp):
        token = resp.json()
        token['token_type'] = 'Bearer'
        resp._content = to_unicode(json.dumps(token)).encode('utf-8')
        return resp
    session.register_compliance_hook('access_token_response', _fix)


class Slack(OAuthBackend):
    OAUTH_TYPE = '2.0'
    OAUTH_NAME = 'slack'
    # Authlib will always add a client_id in payload
    # When use "client_secret_basic" Slack will report error
    OAUTH_CONFIG = {
        'api_base_url': 'https://slack.com/api/',
        'access_token_url': 'https://slack.com/api/oauth.access',
        'authorize_url': 'https://slack.com/oauth/authorize',
        'client_kwargs': {
            'token_endpoint_auth_method': 'client_secret_post',
            'scope': 'identity.basic identity.avatar identity.email',
        },
        'compliance_fix': slack_compliance_fix
    }

    def profile(self, **kwargs):
        resp = self.get('users.identity', **kwargs)
        data = resp.json()
        resp.raise_for_status()
        user = data['user']

        picture = None
        for s in ['512', '192', '72', '48', '32', '24']:
            src = user.get('image_' + s)
            if src:
                picture = src
                break

        params = {
            'sub': user['id'],
            'email': user['email'],
            'name': user['name'],
        }
        if picture:
            params['picture'] = picture
        return UserInfo(params)



#!/usr/bin/env python
import re
from setuptools import setup

_version_re = re.compile(r"version\s=\s'(.*)'")


with open('loginpass/_consts.py', 'r') as f:
    version = _version_re.search(f.read()).group(1)


with open('README.rst') as read_me:
    long_description = read_me.read()


setup(
    name='loginpass',
    version=version,
    description='Social connections powered by Authlib for Flask and Django',
    long_description=long_description,
    url='https://authlib.org/',
    zip_safe=False,
    license='AGPLv3+',
    packages=['loginpass'],
    install_requires=['Authlib'],
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



from ._core import register_to
from ._core import create_flask_blueprint, create_django_urlpatterns
from ._consts import version, homepage
from .google import Google, GoogleServiceAccount
from .github import GitHub
from .facebook import Facebook
from .twitter import Twitter
from .dropbox import Dropbox
from .linkedin import LinkedIn
from .reddit import Reddit
from .gitlab import Gitlab, create_gitlab_backend
from .slack import Slack
from .discord import Discord
from .bitbucket import Bitbucket
from .stackapps import StackOverflow, create_stackapps_backend

__all__ = [
    'register_to',
    'create_flask_blueprint',
    'create_django_urlpatterns',
    'Google', 'GoogleServiceAccount', 'GitHub',
    'Facebook', 'Twitter', 'Dropbox', 'LinkedIn',
    'Reddit', 'Gitlab', 'create_gitlab_backend',
    'Slack', 'Discord', 'Bitbucket',
    'StackOverflow', 'create_stackapps_backend',
]

__version__ = version
__homepage__ = homepage



"""
    loginpass.github
    ~~~~~~~~~~~~~~~~

    Loginpass Backend of GitHub (https://github.com).

    Useful Links:

    - Create App: https://github.com/settings/developers
    - API documentation: https://developer.github.com/v3/

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

import time
from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend


class GitHub(OAuthBackend):
    OAUTH_TYPE = '2.0'
    OAUTH_NAME = 'github'
    OAUTH_CONFIG = {
        'api_base_url': 'https://api.github.com/',
        'access_token_url': 'https://github.com/login/oauth/access_token',
        'authorize_url': 'https://github.com/login/oauth/authorize',
        'client_kwargs': {'scope': 'user:email'},
    }

    def profile(self, **kwargs):
        resp = self.get('user', **kwargs)
        resp.raise_for_status()
        data = resp.json()
        params = {
            'sub': str(data['id']),
            'name': data['name'],
            'email': data.get('email'),
            'preferred_username': data['login'],
            'profile': data['html_url'],
            'picture': data['avatar_url'],
            'website': data.get('blog'),
        }
        updated_at = data.get('updated_at')
        if updated_at:
            t = time.strptime(updated_at, '%Y-%m-%dT%H:%M:%SZ')
            params['updated_at'] = int(time.mktime(t))
        return UserInfo(params)



"""
    loginpass.reddit
    ~~~~~~~~~~~~~~~~

    Loginpass Backend of Reddit (https://www.reddit.com).

    Useful Links:

    - Create App: https://www.reddit.com/prefs/apps
    - API documentation: http://www.reddit.com/dev/api

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend, map_profile_fields
from ._consts import version, homepage


# see: https://github.com/reddit-archive/reddit/wiki/API#rules
UA = 'web:loginpass:{} (by /u/lepture) (+{})'.format(version, homepage)


class Reddit(OAuthBackend):
    OAUTH_TYPE = '2.0'
    OAUTH_NAME = 'reddit'
    OAUTH_CONFIG = {
        'api_base_url': 'https://oauth.reddit.com/api/v1/',
        'access_token_url': 'https://www.reddit.com/api/v1/access_token',
        'authorize_url': 'https://www.reddit.com/api/v1/authorize',
        'client_kwargs': {'scope': 'identity'},
    }

    DEFAULT_USER_AGENT = UA

    def profile(self, **kwargs):
        resp = self.get('me', **kwargs)
        resp.raise_for_status()
        return UserInfo(map_profile_fields(resp.json(), {
            'sub': 'id',
            'name': 'name',
            'email': 'email',
            'preferred_username': 'name',
            'profile': _get_profile,
            'picture': 'icon_img',
            'email_verified': 'has_verified_email',
        }))


def _get_profile(data):
    return 'https://www.reddit.com/user/{}/'.format(data['name'])



"""
    loginpass.gitlab
    ~~~~~~~~~~~~~~~~

    Loginpass Backend of Gitlab (https://gitlab.com) and its
    enterprise endpoints.

    Useful Links:

    - API documentation: https://docs.gitlab.com/ee/api/

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend


def create_gitlab_backend(name, hostname):
    """Build Gitlab OAuth Backend."""
    api_base_url = 'https://{hostname}/api/v4/'.format(hostname=hostname)
    authorize_url = 'https://{hostname}/oauth/authorize'.format(hostname=hostname)
    token_url = 'https://{hostname}/oauth/token'.format(hostname=hostname)

    class GitlabEE(OAuthBackend):
        OAUTH_TYPE = '2.0'
        OAUTH_NAME = name
        OAUTH_CONFIG = {
            'api_base_url': api_base_url,
            'access_token_url': token_url,
            'authorize_url': authorize_url,
            'client_kwargs': {'scope': 'read_user'},
        }

        def profile(self, **kwargs):
            resp = self.get('user', **kwargs)
            resp.raise_for_status()
            data = resp.json()
            params = {
                'sub': str(data['id']),
                'name': data['name'],
                'email': data.get('email'),
                'preferred_username': data['username'],
                'profile': data['web_url'],
                'picture': data['avatar_url'],
                'website': data.get('website_url'),
            }
            return UserInfo(params)

        def parse_openid(self, token, nonce=None):
            # https://docs.gitlab.com/ee/integration/openid_connect_provider.html
            # Although Gitlab claims that it support OIDC, the return
            # value doesn't contains the information it claims.
            raise NotImplementedError()

    return GitlabEE


Gitlab = create_gitlab_backend('gitlab', 'gitlab.com')



"""
    loginpass.linkedin
    ~~~~~~~~~~~~~~~~~~

    Loginpass Backend of LinkedIn (https://linkedin.com).

    Useful Links:

    - API documentation: https://developer.linkedin.com/

    :copyright: (c) 2018 by Hsiaoming Yang
    :license: AGPLv3+, see LICENSE for more details.
"""

from authlib.specs.oidc import UserInfo
from ._core import OAuthBackend, map_profile_fields


class LinkedIn(OAuthBackend):
    OAUTH_TYPE = '2.0'
    OAUTH_NAME = 'linkedin'
    OAUTH_CONFIG = {
        'api_base_url': 'https://api.linkedin.com/v1/',
        'access_token_url': 'https://www.linkedin.com/oauth/v2/accessToken',
        'authorize_url': 'https://www.linkedin.com/oauth/v2/authorization',
        'client_kwargs': {'scope': 'r_basicprofile r_emailaddress'},
    }

    def profile(self, **kwargs):
        fields = [
            'id', 'email-address', 'picture-url', 'public-profile-url',
            'formatted-name', 'first-name', 'last-name', 'maiden-name',
        ]
        url = 'people/~:({})'.format(','.join(fields))
        resp = self.get(url, **kwargs)
        resp.raise_for_status()
        return UserInfo(map_profile_fields(resp.json(), {
            'sub': 'id',
            'email': 'emailAddress',
            'name': 'formattedName',
            'given_name': 'firstName',
            'family_name': 'lastName',
            'middle_name': 'maidenName',
            'picture': 'pictureUrl',
            'profile': 'publicProfileUrl',
        }))
